package gamehost;

import bank.onto.BankingOntology;
import bank.onto.Credit;
import bank.onto.Debit;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.FSMBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import java.util.Vector;
import strategy.Strategy;
import strategy.onto.GamingOntology;

/** Abstrakte Klasse fr einen Agenten, der es Spielern erlaubt ein Spiel zu Spielen */
public abstract class AbsGameHost extends Agent {
  /**
   * Sollte von der <CODE>onEnd()</CODE>-Methode des <CODE>evaluateMoveBehaviour</CODE>
   * zurckgegeben werden, falls das Spiel beendet ist.
   */
  public static final int GAME_OVER = 0;

  /**
   * Sollte von der <CODE>onEnd()</CODE>-Methode des <CODE>evaluateMoveBehaviour</CODE>
   * zurckgegeben werden, falls das Spiel <I>nicht</I> beendet ist.
   */
  public static final int GAME_NOT_OVER = 1;
  // ***** Verhalten, das dem Spieler zum Ziehen geschickt wird *****

  /**
   * Verhalten, das die Spieler bertragen wird. Es sollte auf eine Zugaufforderung
   * oder eine Nachricht, die das Ende des Spiels signalisiert reagieren. Im Konstruktor zu setzen.
   */
  protected Strategy[] makeMoveBehaviour;
  // ***** Verhalten, mit dem der GameHost Zge auswertet

  /**
   * Verhalten, welches den aktuellen Spieler zum Zug auffordern und dann den Zug des Spielers empfangen und auswerten sollte.
   * Im Konstruktor zu setzen.
   */
  protected Behaviour evaluateMoveBehaviour;
  // ***** nicht-spiel-spezifische Variablen *****

  /** Maximale Anzahl der Mitspieler. Im Konstruktor zu setzen. */
  protected int maxPlayers;
  protected Vector players = new Vector();
  protected int gamesPlayed = 0;

  /** Maximale Anzahl zu spielender Spiele. Im Konstruktor zu setzen. */
  protected int maxGames;
  protected boolean playing = false;

  /** Nummer des aktuellen Spielers. */
  protected int playerZaehler = 0;
  protected AID bank;
  private AID winner;
   // im Konstruktor setzen!
  protected String game;
  protected int fee; //Preis fr ein Spiel
  protected int prize; //fr den Gewinner...

  /**
   * Hier festzulegen: <CODE>makeMoveBehaviour</CODE>, <CODE>evaluateMoveBehaviour</CODE>,
   * <CODE>maxPlayers</CODE>, <CODE>maxGames</CODE>
   */
  public AbsGameHost() {
    super();
    resetGame();
  }

  protected void setup() {
    getContentManager().registerLanguage(new SLCodec());
    getContentManager().registerOntology(GamingOntology.getInstance());
    getContentManager().registerOntology(BankingOntology.getInstance());
    // ***** beim DF registrieren *****
    // DFAgentDescription und ServiceDescription zusammenstellen
    DFAgentDescription dfad = new DFAgentDescription();
    ServiceDescription sd = new ServiceDescription();
    sd.setName(getName());
    sd.addOntology(GamingOntology.ONTOLOGY_NAME);
    sd.setType("HOST");
    dfad.setName(getAID());
    dfad.addServices(sd);
    dfad.addOntology(GamingOntology.ONTOLOGY_NAME);
    // Registrierung senden
    try {
      DFService.register(this, dfad);
    } catch (FIPAException e) {
      System.err.println(getLocalName() + ": registration with DF failed.");
      doDelete();
    }
    // ***** Bank suchen *****
    dfad = new DFAgentDescription();
    sd = new ServiceDescription();
    sd.addOntology(BankingOntology.ONTOLOGY_NAME);
    sd.setType("BANK");
    dfad.addServices(sd);
    dfad.addOntology(BankingOntology.ONTOLOGY_NAME);
    try {
      bank = DFService.search(this, dfad) [0].getName();
    } catch (FIPAException e) {
      System.err.println(getLocalName() + ": No bank found.");
      doDelete();
    }
    // ***** Strategien anmelden
    dfad = new DFAgentDescription();
    sd = new ServiceDescription();
    sd.addOntology(GamingOntology.ONTOLOGY_NAME);
    sd.setType("REPOSITORY");
    dfad.addServices(sd);
    dfad.addOntology(GamingOntology.ONTOLOGY_NAME);
    AID rep = null;
    try {
      rep = DFService.search(this, dfad) [0].getName();
    } catch (FIPAException e) {
      System.err.println(getLocalName() + ": No repository found.");
      doDelete();
    }
    ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
    msg.setOntology(GamingOntology.ONTOLOGY_NAME);
    msg.addReceiver(rep);
    for (int i = 0; i < makeMoveBehaviour.length; i++) {
      try {
        msg.setContentObject(makeMoveBehaviour[i]);
        send(msg);
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    // ***** Verhalten hinzufgen *****
    addBehaviour(new ReceivePlayer(this));
    addBehaviour(new ReceiveDeregister(this));
  }

  public void takeDown() {
    try {
      DFService.deregister(this);
    } catch (FIPAException e) {
      System.err.println(getLocalName() + ": error while deregistering.");
    }
  }

  /**
   * Gibt an, ob bereits die maximale Anzahl von Spielern angemeldet ist.
   * @return <CODE>true</CODE>, falls die maximale Anzahl von Spielern angemeldet ist, sonst <CODE>false</CODE>
   */
  public boolean full() {
    return (players.size() >= maxPlayers);
  }

  /**
   * Gibt den Vector mit den AIDs der Spieler zurck.
   * @return Vector mit den AIDs der Spieler
   */
  public Vector getPlayers() { return players; }

  /**
   * Prft, ob ein Spieler angemeldet ist.
   * @param p Spieler, dessen Anmeldung geprft werden soll.
   * @return <CODE>true</CODE>, falls <CODE>p</CODE> angemeldet, sonst <CODE>false</CODE>
   */
  public boolean isPlayer(AID p) {
    return players.contains(p);
  }

  /**
   * Versucht, einen Spieler aufzunehmen.
   * @param p Spieler, der aufgenommen werden soll.
   * @return <CODE>true</CODE>, falls Aufnahme erfolgreich, sonst <CODE>false</CODE>
   */
  public boolean addPlayer(AID p) {
    if (!full()) {
      players.addElement(p);
      return true;
    } else
      return false;
  }

  /**
   * Versucht, einen Spieler aus der Liste der Mitspieler zu entfernen.
   * @param p Spieler, der entfernt werden soll.
   * @return <CODE>true</CODE>, falls Entfernung erfolgreich, sonst <CODE>false</CODE>
   */
  public boolean deletePlayer(AID p) {
    boolean b = players.removeElement(p);
    if (b) {
      ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
      msg.setSender(getAID());
      msg.setContent("Lost");
      msg.addReceiver(p);
      send(msg);
    }
    return b;
  }

  /**
   * Prft, ob bereits die maximale Anzahl von Spielen gespielt wurde.
   * @return <CODE>true</CODE>, falls die maximale Anzahl von Spielen gespielt wurde, sonst <CODE>false</CODE>
   */
  public boolean maxGamesPlayed() { return (gamesPlayed >= maxGames); }

  /** Erhht die Anzahl gespielter Spiele um eins. */
  private void incrementGamesPlayed() { gamesPlayed++; }

  /**
   * Prft, ob gerade ein Spiel luft.
   * @return <CODE>true</CODE>, falls ein Spiel luft, sonst <CODE>false</CODE>
   */
  public boolean isPlaying() { return playing; }

  /** Startet das Spiel. */
  public void startPlaying() {
    playing = true;
    sendStart();
    int num = players.size();
    for (int i = 0; i < num; i++) collectFee((AID)players.elementAt(i));
    // Spielverhalten i.e.S. hinzufgen
    addBehaviour(constructFSM());
  }

  /** Beendet das Spiel. */
  private void stopPlaying() {
    playing = false;
  }

  private FSMBehaviour constructFSM() {
    FSMBehaviour fsm = new FSMBehaviour(this);
    // Zustnde einfgen
    fsm.registerFirstState(evaluateMoveBehaviour, "EvaluateMove");
    fsm.registerLastState(new RemovePlayer(this), "GameOver");
    // Transitionen einfgen
    fsm.registerTransition("EvaluateMove", "GameOver", GAME_OVER);
    fsm.registerTransition("EvaluateMove", "EvaluateMove", GAME_NOT_OVER);
    return fsm;
  }

  /**
   * Gibt die Nummer des aktuellen Spielers zurck.
   * @return Nummer des aktuellen Spielers
   */
  public int getPlayerZaehler() { return playerZaehler; }

  /**
   * Gibt die AID des aktuellen Spielers zurck.
   * @return AID des aktuellen Spielers
   */
  public AID getCurrentPlayer() { return (AID)players.elementAt(playerZaehler); }

  /**
   * Erhht die Nummer des aktuellen Spielers um eins.
   * Falls der aktuelle Spieler der letzte Spieler war, wird sie auf null gesetzt.
   */
  public void incrementPlayerZaehler() {
    if (playerZaehler < (players.size() - 1))
      playerZaehler++; // nchster Spieler ist dran
    else {
      playerZaehler = 0;
    }
  }

  /** Entfernt alle Spieler, erhht die Anzahl gespielter Spiel um eins und beendet das Spiel. */
  public void finishGame() {
    incrementGamesPlayed();
    players.removeAllElements();
    stopPlaying();
  }

  /** Sollte berschrieben werden, um spiel-spezifische Felder zurckzusetzen. */
  public void resetGame() {
    winner = null;
    playerZaehler = 0;
  }

  public AID getBank() {
    return bank;
  }

  public String getGame() {
    return game;
  }

  public int getFee() {
    return fee;
  }

  public int getPrize() {
    return prize;
  }

  public void sendPrize(AID player) {
    ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
    msg.setLanguage(FIPANames.ContentLanguage.FIPA_SL);
    msg.setOntology(BankingOntology.ONTOLOGY_NAME);
    msg.addReceiver(bank);
    Credit c = new Credit();
    c.setOwner(player);
    c.setAmount(prize);
    Action action = new Action();
    action.setActor(bank);
    action.setAction(c);
    try {
      getContentManager().fillContent(msg, action);
      send(msg);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  private void collectFee(AID player) {
    ACLMessage bankMsg = new ACLMessage(ACLMessage.REQUEST);
    bankMsg.setLanguage(FIPANames.ContentLanguage.FIPA_SL);
    bankMsg.setOntology(BankingOntology.ONTOLOGY_NAME);
    bankMsg.setSender(getAID());
    bankMsg.addReceiver(bank);
    Debit d = new Debit();
    d.setOwner(player);
    d.setAmount(fee);
    Action action = new Action();
    action.setActor(bank);
    action.setAction(d);
    try {
      getContentManager().fillContent(bankMsg, action);
      send(bankMsg);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public AID getWinner() {
    return winner;
  }

  public void setWinner(AID winner) {
    this.winner = winner;
  }

  private void sendStart() {
    ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
    msg.setSender(getAID());
    int num = players.size();
    for (int i = 0; i < num; i++)
      msg.addReceiver((AID)players.elementAt(i));
    msg.setContent("Start");
    send(msg);
  }
}
